# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import openstack
from openstack.connection import exceptions as openstack_exc

from ironic.common import exception
from ironic.common import keystone
from ironic.common import swift
from ironic.conf import CONF
from ironic.tests import base


@mock.patch.object(swift, 'get_swift_session', autospec=True,
                   return_value=mock.MagicMock(verify=False,
                                               cert=('spam', 'ham'),
                                               timeout=42))
@mock.patch.object(openstack.connection, 'Connection', autospec=True)
class SwiftTestCase(base.TestCase):

    def setUp(self):
        super(SwiftTestCase, self).setUp()
        self.swift_exception = openstack_exc.OpenStackCloudException()

    @mock.patch.object(keystone, 'get_auth', autospec=True,
                       return_value=mock.Mock())
    @mock.patch.object(keystone, 'get_endpoint', autospec=True,
                       return_value='http://example.com/v1')
    def test___init__(self, get_endpoint_mock, get_auth_mock, connection_mock,
                      session_mock):
        """Check if client is properly initialized with swift"""
        swiftapi = swift.SwiftAPI()
        connection_mock.assert_called_once_with(
            session=session_mock.return_value,
            oslo_conf=CONF
        )
        self.assertEqual(connection_mock.return_value, swiftapi.connection)

    def test_create_object(self, connection_mock, session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        connection.create_object.return_value = 'object-uuid'

        object_uuid = swiftapi.create_object('container', 'object',
                                             'some-file-location',
                                             object_headers={'foo': 'bar'})

        connection.create_container.assert_called_once_with('container')
        connection.create_object.assert_called_once_with(
            'container', 'object', filename='some-file-location', foo='bar')
        self.assertEqual('object-uuid', object_uuid)

    def test_create_object_create_container_fails(self, connection_mock,
                                                  session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        connection.create_container.side_effect = self.swift_exception
        self.assertRaises(exception.SwiftOperationError,
                          swiftapi.create_object, 'container',
                          'object', 'some-file-location')
        connection.create_container.assert_called_once_with('container')
        self.assertFalse(connection.create_object.called)

    def test_create_object_create_object_fails(self, connection_mock,
                                               session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        connection.create_object.side_effect = self.swift_exception
        self.assertRaises(exception.SwiftOperationError,
                          swiftapi.create_object, 'container',
                          'object', 'some-file-location')
        connection.create_container.assert_called_once_with('container')
        connection.create_object.assert_called_once_with(
            'container', 'object', filename='some-file-location')

    def test_create_object_from_data(self, connection_mock, session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        connection.create_object.return_value = 'object-uuid'

        object_uuid = swiftapi.create_object_from_data(
            'object', 'some-data', 'container')

        connection.create_container.assert_called_once_with('container')
        connection.create_object.assert_called_once_with(
            'container', 'object', data='some-data')
        self.assertEqual('object-uuid', object_uuid)

    def test_create_object_from_data_create_container_fails(
            self, connection_mock, session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        connection.create_container.side_effect = self.swift_exception
        self.assertRaises(exception.SwiftOperationError,
                          swiftapi.create_object_from_data,
                          'object', 'some-data', 'container')
        connection.create_container.assert_called_once_with('container')
        self.assertFalse(connection.create_object.called)

    def test_create_object_from_data_create_object_fails(self, connection_mock,
                                                         session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        connection.create_object.side_effect = self.swift_exception
        self.assertRaises(exception.SwiftOperationError,
                          swiftapi.create_object_from_data,
                          'object', 'some-data', 'container')
        connection.create_container.assert_called_once_with('container')
        connection.create_object.assert_called_once_with(
            'container', 'object', data='some-data')

    @mock.patch.object(keystone, 'get_auth', autospec=True)
    @mock.patch.object(keystone, 'get_endpoint', autospec=True)
    def test_get_temp_url(self, get_endpoint_mock, get_auth_mock,
                          connection_mock, session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        get_endpoint_mock.return_value = 'http://example.com/v1/AUTH_tenant_id'
        object_store = connection.object_store
        object_store.get_temp_url_key.return_value = 'secretkey'

        object_store.generate_temp_url.return_value = \
            '/v1/AUTH_tenant_id/temp-url-path'
        temp_url_returned = swiftapi.get_temp_url('container', 'object', 10)
        object_path_expected = '/v1/AUTH_tenant_id/container/object'
        object_store.generate_temp_url.assert_called_once_with(
            object_path_expected, 10, 'GET', temp_url_key='secretkey')
        self.assertEqual('http://example.com/v1/AUTH_tenant_id/temp-url-path',
                         temp_url_returned)

    def test_delete_object(self, connection_mock, session_mock):
        swiftapi = swift.SwiftAPI()
        connection = connection_mock.return_value
        swiftapi.delete_object('container', 'object')
        connection.delete_object.assert_called_once_with('container',
                                                         'object')

    def test_delete_object_exc_resource_not_found(self, connection_mock,
                                                  session_mock):
        swiftapi = swift.SwiftAPI()
        exc = openstack_exc.ResourceNotFound(message="Resource not found")
        connection = connection_mock.return_value
        connection.delete_object.side_effect = exc
        self.assertRaises(exception.SwiftObjectNotFoundError,
                          swiftapi.delete_object, 'container', 'object')
        connection.delete_object.assert_called_once_with('container',
                                                         'object')

    def test_delete_object_exc(self, connection_mock, session_mock):
        swiftapi = swift.SwiftAPI()
        exc = self.swift_exception
        connection = connection_mock.return_value
        connection.delete_object.side_effect = exc
        self.assertRaises(exception.SwiftOperationError,
                          swiftapi.delete_object, 'container', 'object')
        connection.delete_object.assert_called_once_with('container',
                                                         'object')



# Copyright 2010-2011 OpenStack Foundation
# Copyright 2012-2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for database migrations. There are "opportunistic" tests for mysql in
here, which allow testing against a database in a properly configured unit
test environment.

For the opportunistic testing you need to set up a db named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost.
The test will then use that db and u/p combo to run the tests. See
tools/test-setup.sh in the root of the repo for a full setup example.
"""

import collections
import contextlib
import json
import os
from unittest import mock

from alembic import script
import fixtures
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log import log as logging
from oslo_utils import uuidutils
from oslotest import base as test_base
import sqlalchemy
import sqlalchemy.exc

from ironic.conf import CONF
from ironic.db.sqlalchemy import migration
from ironic.db.sqlalchemy import models
from ironic.tests import base

LOG = logging.getLogger(__name__)

# NOTE(vdrok): This was introduced after migration tests started taking more
# time in gate. Timeout value in seconds for tests performing migrations.
# Can be modified in tox.ini as env variable.
MIGRATIONS_TIMEOUT = os.getenv('MIGRATIONS_TIMEOUT', 60)


@contextlib.contextmanager
def patch_with_engine(engine):
    with mock.patch.object(enginefacade.writer, 'get_engine',
                           autospec=True) as patch_engine:
        patch_engine.return_value = engine
        yield


class WalkVersionsMixin(object):
    def _walk_versions(self, engine=None, alembic_cfg=None):
        # Determine latest version script from the repo, then
        # upgrade from 1 through to the latest, with no data
        # in the databases. This just checks that the schema itself
        # upgrades successfully.

        # Place the database under version control
        with patch_with_engine(engine):

            script_directory = script.ScriptDirectory.from_config(alembic_cfg)

            self.assertIsNone(self.migration_api.version(alembic_cfg))

            versions = [ver for ver in script_directory.walk_revisions()]

            for version in reversed(versions):
                self._migrate_up(engine, alembic_cfg,
                                 version.revision, with_data=True)

    def _migrate_up(self, engine, config, version, with_data=False):
        """migrate up to a new version of the db.

        We allow for data insertion and post checks at every
        migration version with special _pre_upgrade_### and
        _check_### functions in the main test.
        """
        # NOTE(sdague): try block is here because it's impossible to debug
        # where a failed data migration happens otherwise
        try:
            if with_data:
                data = None
                pre_upgrade = getattr(
                    self, "_pre_upgrade_%s" % version, None)
                if pre_upgrade:
                    data = pre_upgrade(engine)

            self.migration_api.upgrade(version, config=config)
            self.assertEqual(version, self.migration_api.version(config))
            if with_data:
                check = getattr(self, "_check_%s" % version, None)
                if check:
                    check(engine, data)

        except Exception:
            LOG.error("Failed to migrate to version %(version)s on engine "
                      "%(engine)s",
                      {'version': version, 'engine': engine})
            raise


class TestWalkVersions(base.TestCase, WalkVersionsMixin):
    def setUp(self):
        super(TestWalkVersions, self).setUp()
        self.migration_api = mock.MagicMock()
        self.engine = mock.MagicMock()
        self.config = mock.MagicMock()
        self.versions = [mock.Mock(revision='2b2'), mock.Mock(revision='1a1')]

    def test_migrate_up(self):
        self.migration_api.version.return_value = 'dsa123'

        self._migrate_up(self.engine, self.config, 'dsa123')

        self.migration_api.upgrade.assert_called_with('dsa123',
                                                      config=self.config)
        self.migration_api.version.assert_called_with(self.config)

    def test_migrate_up_with_data(self):
        test_value = {"a": 1, "b": 2}
        self.migration_api.version.return_value = '141'
        self._pre_upgrade_141 = mock.MagicMock()
        self._pre_upgrade_141.return_value = test_value
        self._check_141 = mock.MagicMock()

        self._migrate_up(self.engine, self.config, '141', True)

        self._pre_upgrade_141.assert_called_with(self.engine)
        self._check_141.assert_called_with(self.engine, test_value)

    @mock.patch.object(script, 'ScriptDirectory', autospec=True)
    @mock.patch.object(WalkVersionsMixin, '_migrate_up', autospec=True)
    def test_walk_versions_all_default(self, _migrate_up, script_directory):
        fc = script_directory.from_config.return_value
        fc.walk_revisions.return_value = self.versions
        self.migration_api.version.return_value = None

        self._walk_versions(self.engine, self.config)

        self.migration_api.version.assert_called_with(self.config)

        upgraded = [mock.call(self, self.engine, self.config, v.revision,
                    with_data=True) for v in reversed(self.versions)]
        self.assertEqual(self._migrate_up.call_args_list, upgraded)

    @mock.patch.object(script, 'ScriptDirectory', autospec=True)
    @mock.patch.object(WalkVersionsMixin, '_migrate_up', autospec=True)
    def test_walk_versions_all_false(self, _migrate_up, script_directory):
        fc = script_directory.from_config.return_value
        fc.walk_revisions.return_value = self.versions
        self.migration_api.version.return_value = None

        self._walk_versions(self.engine, self.config)

        upgraded = [mock.call(self, self.engine, self.config, v.revision,
                    with_data=True) for v in reversed(self.versions)]
        self.assertEqual(upgraded, self._migrate_up.call_args_list)


class MigrationCheckersMixin(object):

    def setUp(self):
        super(MigrationCheckersMixin, self).setUp()
        self.engine = enginefacade.writer.get_engine()
        self.config = migration._alembic_config()
        self.migration_api = migration
        self.useFixture(fixtures.Timeout(int(MIGRATIONS_TIMEOUT),
                                         gentle=False))

    def test_walk_versions(self):
        self._walk_versions(self.engine, self.config)

    def _check_21b331f883ef(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('provision_updated_at', col_names)
        self.assertIsInstance(nodes.c.provision_updated_at.type,
                              sqlalchemy.types.DateTime)

    def _check_3cb628139ea4(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]

        self.assertIn('console_enabled', col_names)
        # in some backends bool type is integer
        self.assertIsInstance(nodes.c.console_enabled.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))

    def _check_31baaf680d2b(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('instance_info', col_names)
        self.assertIsInstance(nodes.c.instance_info.type,
                              sqlalchemy.types.TEXT)

    def _check_3bea56f25597(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        instance_uuid = 'aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee'
        data = {'driver': 'fake',
                'uuid': uuidutils.generate_uuid(),
                'instance_uuid': instance_uuid}

        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            data['uuid'] = uuidutils.generate_uuid()
            self.assertRaises(db_exc.DBDuplicateEntry,
                              connection.execute, nodes.insert(), data)

    def _check_487deb87cc9d(self, engine, data):
        conductors = db_utils.get_table(engine, 'conductors')
        column_names = [column.name for column in conductors.c]

        self.assertIn('online', column_names)
        self.assertIsInstance(conductors.c.online.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))
        nodes = db_utils.get_table(engine, 'nodes')
        column_names = [column.name for column in nodes.c]
        self.assertIn('conductor_affinity', column_names)
        self.assertIsInstance(nodes.c.conductor_affinity.type,
                              sqlalchemy.types.Integer)

        data_conductor = {'hostname': 'test_host'}
        with engine.begin() as connection:
            insert_conductor = conductors.insert().values(data_conductor)
            connection.execute(insert_conductor)
            conductor_stmt = sqlalchemy.select(
                models.Conductor.id
            ).where(
                models.Conductor.hostname == 'test_host'
            )
            conductor = connection.execute(conductor_stmt).first()
            data_node = {'uuid': uuidutils.generate_uuid(),
                         'conductor_affinity': conductor.id}
            insert_node = nodes.insert().values(data_node)

            connection.execute(insert_node)
            node_stmt = sqlalchemy.select(
                models.Node.conductor_affinity
            ).where(
                models.Node.uuid == data_node['uuid']
            )
            node = connection.execute(node_stmt).first()
            self.assertEqual(conductor.id, node.conductor_affinity)

    def _check_242cc6a923b3(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('maintenance_reason', col_names)
        self.assertIsInstance(nodes.c.maintenance_reason.type,
                              sqlalchemy.types.String)

    def _pre_upgrade_5674c57409b9(self, engine):
        # add some nodes in various states so we can assert that "None"
        # was replaced by "available", and nothing else changed.
        nodes = db_utils.get_table(engine, 'nodes')
        data = [{'uuid': uuidutils.generate_uuid(),
                 'provision_state': 'fake state'},
                {'uuid': uuidutils.generate_uuid(),
                 'provision_state': 'active'},
                {'uuid': uuidutils.generate_uuid(),
                 'provision_state': 'deleting'},
                {'uuid': uuidutils.generate_uuid(),
                 'provision_state': None}]
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
        return data

    def _check_5674c57409b9(self, engine, data):
        with engine.begin() as connection:
            result = connection.execute(
                sqlalchemy.select(
                    models.Node.uuid,
                    models.Node.provision_state
                )
            )

            def _get_state(uuid):
                for row in data:
                    if row['uuid'] == uuid:
                        return row['provision_state']

            for row in result:
                old = _get_state(row.uuid)
                new = row.provision_state
                if old is None:
                    self.assertEqual('available', new)
                else:
                    self.assertEqual(old, new)

    def _check_bb59b63f55a(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('driver_internal_info', col_names)
        self.assertIsInstance(nodes.c.driver_internal_info.type,
                              sqlalchemy.types.TEXT)

    def _check_3ae36a5f5131(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        column_names = [column.name for column in nodes.c]
        self.assertIn('name', column_names)
        self.assertIsInstance(nodes.c.name.type,
                              sqlalchemy.types.String)
        data = {'driver': 'fake',
                'uuid': uuidutils.generate_uuid(),
                'name': 'node'
                }
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            data['uuid'] = uuidutils.generate_uuid()
            self.assertRaises(db_exc.DBDuplicateEntry, connection.execute,
                              nodes.insert(), data)

    def _check_1e1d5ace7dc6(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        column_names = [column.name for column in nodes.c]
        self.assertIn('inspection_started_at', column_names)
        self.assertIn('inspection_finished_at', column_names)
        self.assertIsInstance(nodes.c.inspection_started_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(nodes.c.inspection_finished_at.type,
                              sqlalchemy.types.DateTime)

    def _check_01f21d5e5195(self, engine, data):
        node_history = db_utils.get_table(engine, 'node_history')
        bigstring = 'a' * 64
        uuid = uuidutils.generate_uuid()
        data = {'uuid': uuid, 'user': bigstring}
        with engine.begin() as connection:
            insert_node_history = node_history.insert().values(data)
            connection.execute(insert_node_history)
            node_history_stmt = sqlalchemy.select(
                models.NodeHistory.user
            ).where(
                models.NodeHistory.uuid == uuid
            )
            node_history = connection.execute(node_history_stmt).first()
            self.assertEqual(bigstring, node_history.user)

    def _check_4f399b21ae71(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('clean_step', col_names)
        self.assertIsInstance(nodes.c.clean_step.type,
                              sqlalchemy.types.String)

    def _check_789acc877671(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('raid_config', col_names)
        self.assertIn('target_raid_config', col_names)
        self.assertIsInstance(nodes.c.raid_config.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(nodes.c.target_raid_config.type,
                              sqlalchemy.types.String)

    def _check_2fb93ffd2af1(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        bigstring = 'a' * 255
        uuid = uuidutils.generate_uuid()
        data = {'uuid': uuid, 'name': bigstring}
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            node_stmt = sqlalchemy.select(
                models.Node.name
            ).where(
                models.Node.uuid == uuid
            )
            node = connection.execute(node_stmt).first()
            self.assertEqual(bigstring, node.name)

    def _check_516faf1bb9b1(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        bigstring = 'a' * 255
        uuid = uuidutils.generate_uuid()
        data = {'uuid': uuid, 'driver': bigstring}
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            node_stmt = sqlalchemy.select(
                models.Node.driver
            ).where(
                models.Node.uuid == uuid
            )
            node = connection.execute(node_stmt).first()
            self.assertEqual(bigstring, node.driver)

    def _check_48d6c242bb9b(self, engine, data):
        node_tags = db_utils.get_table(engine, 'node_tags')
        col_names = [column.name for column in node_tags.c]
        self.assertIn('tag', col_names)
        self.assertIsInstance(node_tags.c.tag.type,
                              sqlalchemy.types.String)
        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            data = {'id': '123', 'name': 'node1'}
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            data = {'node_id': '123', 'tag': 'tag1'}
            insert_node_tag = node_tags.insert().values(data)
            connection.execute(insert_node_tag)
            tag_stmt = sqlalchemy.select(
                models.NodeTag.tag
            ).where(
                models.NodeTag.node_id == '123'
            )
            tag = connection.execute(tag_stmt).first()
            self.assertEqual('tag1', tag.tag)

    def _check_5ea1b0d310e(self, engine, data):
        portgroup = db_utils.get_table(engine, 'portgroups')
        col_names = [column.name for column in portgroup.c]
        expected_names = ['created_at', 'updated_at', 'id', 'uuid', 'name',
                          'node_id', 'address', 'extra']
        self.assertEqual(sorted(expected_names), sorted(col_names))

        self.assertIsInstance(portgroup.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(portgroup.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(portgroup.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(portgroup.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(portgroup.c.name.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(portgroup.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(portgroup.c.address.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(portgroup.c.extra.type,
                              sqlalchemy.types.TEXT)

        ports = db_utils.get_table(engine, 'ports')
        col_names = [column.name for column in ports.c]
        self.assertIn('pxe_enabled', col_names)
        self.assertIn('portgroup_id', col_names)
        self.assertIn('local_link_connection', col_names)
        self.assertIsInstance(ports.c.portgroup_id.type,
                              sqlalchemy.types.Integer)
        # in some backends bool type is integer
        self.assertIsInstance(ports.c.pxe_enabled.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))

    def _pre_upgrade_f6fdb920c182(self, engine):
        # add some ports.
        ports = db_utils.get_table(engine, 'ports')
        data = [{'uuid': uuidutils.generate_uuid(), 'pxe_enabled': None},
                {'uuid': uuidutils.generate_uuid(), 'pxe_enabled': None}]
        with engine.begin() as connection:
            insert_ports = ports.insert().values(data)
            connection.execute(insert_ports)
        return data

    def _check_f6fdb920c182(self, engine, data):

        with engine.begin() as connection:
            port_stmt = sqlalchemy.select(
                models.Port.uuid,
                models.Port.pxe_enabled
            )
            result = connection.execute(port_stmt)

            def _was_inserted(uuid):
                for row in data:
                    if row['uuid'] == uuid:
                        return True

            for row in result:
                if _was_inserted(row.uuid):
                    self.assertTrue(row.pxe_enabled)

    def _check_e294876e8028(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('network_interface', col_names)
        self.assertIsInstance(nodes.c.network_interface.type,
                              sqlalchemy.types.String)

    def _check_10b163d4481e(self, engine, data):
        ports = db_utils.get_table(engine, 'ports')
        portgroups = db_utils.get_table(engine, 'portgroups')
        port_col_names = [column.name for column in ports.c]
        portgroup_col_names = [column.name for column in portgroups.c]
        self.assertIn('internal_info', port_col_names)
        self.assertIn('internal_info', portgroup_col_names)
        self.assertIsInstance(ports.c.internal_info.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(portgroups.c.internal_info.type,
                              sqlalchemy.types.TEXT)

    def _check_dd34e1f1303b(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('resource_class', col_names)
        self.assertIsInstance(nodes.c.resource_class.type,
                              sqlalchemy.types.String)

    def _pre_upgrade_c14cef6dfedf(self, engine):
        # add some nodes.
        nodes = db_utils.get_table(engine, 'nodes')
        data = [{'uuid': uuidutils.generate_uuid(),
                 'network_interface': None},
                {'uuid': uuidutils.generate_uuid(),
                 'network_interface': None},
                {'uuid': uuidutils.generate_uuid(),
                 'network_interface': 'neutron'}]
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
        return data

    def _check_c14cef6dfedf(self, engine, data):
        with engine.begin() as connection:
            counts = collections.defaultdict(int)
            result = connection.execute(
                sqlalchemy.select(
                    models.Node.uuid,
                    models.Node.network_interface))

            def _was_inserted(uuid):
                for row in data:
                    if row['uuid'] == uuid:
                        return True

            for row in result:
                if _was_inserted(row.uuid):
                    counts[row.network_interface] += 1

            # using default conf values, we should have 2 flat and one neutron
            self.assertEqual(2, counts['flat'])
            self.assertEqual(1, counts['neutron'])
            self.assertEqual(0, counts[None])

    def _check_60cf717201bc(self, engine, data):
        portgroups = db_utils.get_table(engine, 'portgroups')
        col_names = [column.name for column in portgroups.c]
        self.assertIn('standalone_ports_supported', col_names)
        self.assertIsInstance(portgroups.c.standalone_ports_supported.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))

    def _check_bcdd431ba0bf(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        added_ifaces = ['boot', 'console', 'deploy', 'inspect',
                        'management', 'power', 'raid', 'vendor']
        for iface in added_ifaces:
            name = '%s_interface' % iface
            self.assertIn(name, col_names)
            self.assertIsInstance(getattr(nodes.c, name).type,
                                  sqlalchemy.types.String)

    def _check_daa1ba02d98(self, engine, data):
        connectors = db_utils.get_table(engine, 'volume_connectors')
        col_names = [column.name for column in connectors.c]
        expected_names = ['created_at', 'updated_at', 'id', 'uuid', 'node_id',
                          'type', 'connector_id', 'extra']
        self.assertEqual(sorted(expected_names), sorted(col_names))

        self.assertIsInstance(connectors.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(connectors.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(connectors.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(connectors.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(connectors.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(connectors.c.type.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(connectors.c.connector_id.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(connectors.c.extra.type,
                              sqlalchemy.types.TEXT)

    def _check_1a59178ebdf6(self, engine, data):
        targets = db_utils.get_table(engine, 'volume_targets')
        col_names = [column.name for column in targets.c]
        expected_names = ['created_at', 'updated_at', 'id', 'uuid', 'node_id',
                          'boot_index', 'extra', 'properties', 'volume_type',
                          'volume_id']
        self.assertEqual(sorted(expected_names), sorted(col_names))

        self.assertIsInstance(targets.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(targets.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(targets.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(targets.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(targets.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(targets.c.boot_index.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(targets.c.extra.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(targets.c.properties.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(targets.c.volume_type.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(targets.c.volume_id.type,
                              sqlalchemy.types.String)

    def _pre_upgrade_493d8f27f235(self, engine):
        portgroups = db_utils.get_table(engine, 'portgroups')
        data = [{'uuid': uuidutils.generate_uuid()},
                {'uuid': uuidutils.generate_uuid()}]
        with engine.begin() as connection:
            insert_portgroups = portgroups.insert().values(data)
            connection.execute(insert_portgroups)
        return data

    def _check_493d8f27f235(self, engine, data):
        portgroups = db_utils.get_table(engine, 'portgroups')
        col_names = [column.name for column in portgroups.c]
        self.assertIn('properties', col_names)
        self.assertIsInstance(portgroups.c.properties.type,
                              sqlalchemy.types.TEXT)
        self.assertIn('mode', col_names)
        self.assertIsInstance(portgroups.c.mode.type,
                              sqlalchemy.types.String)
        with engine.begin() as connection:
            result = connection.execute(
                sqlalchemy.select(models.Portgroup.mode)
            )
            for row in result:
                self.assertEqual(CONF.default_portgroup_mode, row.mode)

    def _check_1d6951876d68(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('storage_interface', col_names)
        self.assertIsInstance(nodes.c.storage_interface.type,
                              sqlalchemy.types.String)

    def _check_2353895ecfae(self, engine, data):
        ifaces = db_utils.get_table(engine, 'conductor_hardware_interfaces')
        col_names = [column.name for column in ifaces.c]
        expected_names = ['created_at', 'updated_at', 'id', 'conductor_id',
                          'hardware_type', 'interface_type', 'interface_name']
        self.assertEqual(sorted(expected_names), sorted(col_names))

        self.assertIsInstance(ifaces.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(ifaces.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(ifaces.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(ifaces.c.conductor_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(ifaces.c.hardware_type.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(ifaces.c.interface_type.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(ifaces.c.interface_name.type,
                              sqlalchemy.types.String)

    def _check_dbefd6bdaa2c(self, engine, data):
        ifaces = db_utils.get_table(engine, 'conductor_hardware_interfaces')
        col_names = [column.name for column in ifaces.c]
        self.assertIn('default', col_names)
        self.assertIsInstance(ifaces.c.default.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))

    def _check_3d86a077a3f2(self, engine, data):
        ports = db_utils.get_table(engine, 'ports')
        col_names = [column.name for column in ports.c]
        self.assertIn('physical_network', col_names)
        self.assertIsInstance(ports.c.physical_network.type,
                              sqlalchemy.types.String)

    def _check_868cb606a74a(self, engine, data):
        for table in ['chassis', 'conductors', 'node_tags', 'nodes',
                      'portgroups', 'ports', 'volume_connectors',
                      'volume_targets', 'conductor_hardware_interfaces']:
            table = db_utils.get_table(engine, table)
            col_names = [column.name for column in table.c]
            self.assertIn('version', col_names)
            self.assertIsInstance(table.c.version.type,
                                  sqlalchemy.types.String)

    def _check_405cfe08f18d(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('rescue_interface', col_names)
        self.assertIsInstance(nodes.c.rescue_interface.type,
                              sqlalchemy.types.String)

    def _pre_upgrade_b4130a7fc904(self, engine):
        # Create a node to which traits can be added.
        data = {'uuid': uuidutils.generate_uuid()}
        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            node_stmt = sqlalchemy.select(
                models.Node.id
            ).where(
                models.Node.uuid == data['uuid']
            )
            node = connection.execute(node_stmt).first()
            # WARNING: Always copy, never directly return a db object or
            # piece of a db object. It is a sqlalchemy thing.
            data['id'] = int(node.id)
        return data

    def _check_b4130a7fc904(self, engine, data):
        node_traits = db_utils.get_table(engine, 'node_traits')
        col_names = [column.name for column in node_traits.c]
        self.assertIn('node_id', col_names)
        self.assertIsInstance(node_traits.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIn('trait', col_names)
        self.assertIsInstance(node_traits.c.trait.type,
                              sqlalchemy.types.String)

        trait = {'node_id': data['id'], 'trait': 'trait1'}
        with engine.begin() as connection:
            insert_trait = node_traits.insert().values(trait)
            connection.execute(insert_trait)
            trait_stmt = sqlalchemy.select(
                models.NodeTrait.trait
            ).where(
                models.NodeTrait.node_id == data['id']
            )
            trait = connection.execute(trait_stmt).first()
            self.assertEqual('trait1', trait.trait)

    def _pre_upgrade_82c315d60161(self, engine):
        # Create a node to which bios setting can be added.
        data = {'uuid': uuidutils.generate_uuid()}
        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            node_stmt = sqlalchemy.select(
                models.Node.id
            ).where(models.Node.uuid == data['uuid'])
            node = connection.execute(node_stmt).first()
            # WARNING: Always copy, never directly return a db object or
            # piece of a db object. It is a sqlalchemy thing.
            data['id'] = int(node.id)
        return data

    def _check_82c315d60161(self, engine, data):
        bios_settings = db_utils.get_table(engine, 'bios_settings')
        col_names = [column.name for column in bios_settings.c]
        expected_names = ['node_id', 'created_at', 'updated_at',
                          'name', 'value', 'version']
        self.assertEqual(sorted(expected_names), sorted(col_names))
        self.assertIsInstance(bios_settings.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(bios_settings.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(bios_settings.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(bios_settings.c.name.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(bios_settings.c.version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(bios_settings.c.value.type,
                              sqlalchemy.types.Text)

        setting = {'node_id': data['id'],
                   'name': 'virtualization',
                   'value': 'on'}
        with engine.begin() as connection:
            insert_bios_settings = bios_settings.insert().values(setting)
            connection.execute(insert_bios_settings)
            setting_stmt = sqlalchemy.select(
                models.BIOSSetting.value
            ).where(
                models.BIOSSetting.node_id == data['id'],
                models.BIOSSetting.name == setting['name']
            )
            setting = connection.execute(setting_stmt).first()
            self.assertEqual('on', setting[0])

    def _check_2bbd96b6ccb9(self, engine, data):
        bios_settings = db_utils.get_table(engine, 'bios_settings')
        col_names = [column.name for column in bios_settings.c]
        self.assertIn('attribute_type', col_names)
        self.assertIn('allowable_values', col_names)
        self.assertIn('lower_bound', col_names)
        self.assertIn('max_length', col_names)
        self.assertIn('min_length', col_names)
        self.assertIn('read_only', col_names)
        self.assertIn('reset_required', col_names)
        self.assertIn('unique', col_names)
        self.assertIn('upper_bound', col_names)
        self.assertIsInstance(bios_settings.c.attribute_type.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(bios_settings.c.allowable_values.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(bios_settings.c.lower_bound.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(bios_settings.c.max_length.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(bios_settings.c.min_length.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(bios_settings.c.read_only.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))
        self.assertIsInstance(bios_settings.c.reset_required.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))
        self.assertIsInstance(bios_settings.c.unique.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))
        self.assertIsInstance(bios_settings.c.upper_bound.type,
                              sqlalchemy.types.Integer)

    def _check_2d13bc3d6bba(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('bios_interface', col_names)
        self.assertIsInstance(nodes.c.bios_interface.type,
                              sqlalchemy.types.String)

    def _check_fb3f10dd262e(self, engine, data):
        nodes_tbl = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes_tbl.c]
        self.assertIn('fault', col_names)
        self.assertIsInstance(nodes_tbl.c.fault.type,
                              sqlalchemy.types.String)

    def _check_b9117ac17882(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('deploy_step', col_names)
        self.assertIsInstance(nodes.c.deploy_step.type,
                              sqlalchemy.types.String)

    def _pre_upgrade_664f85c2f622(self, engine):
        # Create a node and a conductor to verify existing records
        # get a conductor_group of ""
        data = {
            'conductor_id': 98765432,
            'node_uuid': uuidutils.generate_uuid(),
        }

        nodes = db_utils.get_table(engine, 'nodes')
        conductors = db_utils.get_table(engine, 'conductors')
        with engine.begin() as connection:
            insert_node = nodes.insert().values({'uuid': data['node_uuid']})
            connection.execute(insert_node)
            insert_conductor = conductors.insert().values(
                {'id': data['conductor_id'],
                 'hostname': uuidutils.generate_uuid()})
            connection.execute(insert_conductor)
        return data

    def _check_664f85c2f622(self, engine, data):
        nodes_tbl = db_utils.get_table(engine, 'nodes')
        conductors_tbl = db_utils.get_table(engine, 'conductors')
        for tbl in (nodes_tbl, conductors_tbl):
            col_names = [column.name for column in tbl.c]
            self.assertIn('conductor_group', col_names)
            self.assertIsInstance(tbl.c.conductor_group.type,
                                  sqlalchemy.types.String)
        with engine.begin() as connection:
            node_stmt = sqlalchemy.select(
                models.Node.uuid,
                models.Node.conductor_group,
            ).where(
                models.Node.uuid == data['node_uuid'])
            node = connection.execute(node_stmt).first()
            self.assertEqual(node.conductor_group, "")

            conductor_stmt = sqlalchemy.select(
                models.Conductor.conductor_group,
            ).where(
                models.Conductor.id == data['conductor_id'],
            )
            conductor = connection.execute(conductor_stmt).first()
            self.assertEqual(conductor.conductor_group, "")

    def _check_d2b036ae9378(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('automated_clean', col_names)

    def _pre_upgrade_93706939026c(self, engine):
        data = {
            'node_uuid': uuidutils.generate_uuid(),
        }

        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            insert_node = nodes.insert().values({'uuid': data['node_uuid']})
            connection.execute(insert_node)
        return data

    def _check_93706939026c(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('protected', col_names)
        self.assertIn('protected_reason', col_names)

        with engine.begin() as connection:
            node_stmt = sqlalchemy.select(
                models.Node.uuid,
                models.Node.protected,
                models.Node.protected_reason
            ).where(
                models.Node.uuid == data['node_uuid'])
            node = connection.execute(node_stmt).first()
            self.assertFalse(node.protected)
            self.assertIsNone(node.protected_reason)

    def _check_f190f9d00a11(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('owner', col_names)

    def _pre_upgrade_dd67b91a1981(self, engine):
        data = {
            'node_uuid': uuidutils.generate_uuid(),
        }

        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            insert_node = nodes.insert().values({'uuid': data['node_uuid']})
            connection.execute(insert_node)
        return data

    def _check_dd67b91a1981(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('allocation_id', col_names)

        with engine.begin() as connection:
            node_stmt = sqlalchemy.select(
                models.Node.allocation_id
            ).where(
                models.Node.uuid == data['node_uuid']
            )
            node = connection.execute(node_stmt).first()
            self.assertIsNone(node.allocation_id)

        allocations = db_utils.get_table(engine, 'allocations')
        col_names = [column.name for column in allocations.c]
        expected_names = ['id', 'uuid', 'node_id', 'created_at', 'updated_at',
                          'name', 'version', 'state', 'last_error',
                          'resource_class', 'traits', 'candidate_nodes',
                          'extra', 'conductor_affinity']
        self.assertEqual(sorted(expected_names), sorted(col_names))
        self.assertIsInstance(allocations.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(allocations.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(allocations.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(allocations.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(allocations.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(allocations.c.state.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(allocations.c.last_error.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(allocations.c.resource_class.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(allocations.c.traits.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(allocations.c.candidate_nodes.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(allocations.c.extra.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(allocations.c.conductor_affinity.type,
                              sqlalchemy.types.Integer)

    def _check_9cbeefa3763f(self, engine, data):
        ports = db_utils.get_table(engine, 'ports')
        col_names = [column.name for column in ports.c]
        self.assertIn('is_smartnic', col_names)
        # in some backends bool type is integer
        self.assertIsInstance(ports.c.is_smartnic.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))

    def _check_28c44432c9c3(self, engine, data):
        nodes_tbl = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes_tbl.c]
        self.assertIn('description', col_names)
        self.assertIsInstance(nodes_tbl.c.description.type,
                              sqlalchemy.types.TEXT)

    def _check_2aac7e0872f6(self, engine, data):
        # Deploy templates.
        deploy_templates = db_utils.get_table(engine, 'deploy_templates')
        col_names = [column.name for column in deploy_templates.c]
        expected = ['created_at', 'updated_at', 'version',
                    'id', 'uuid', 'name']
        self.assertEqual(sorted(expected), sorted(col_names))
        self.assertIsInstance(deploy_templates.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(deploy_templates.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(deploy_templates.c.version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(deploy_templates.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(deploy_templates.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(deploy_templates.c.name.type,
                              sqlalchemy.types.String)

        # Deploy template steps.
        deploy_template_steps = db_utils.get_table(engine,
                                                   'deploy_template_steps')
        col_names = [column.name for column in deploy_template_steps.c]
        expected = ['created_at', 'updated_at', 'version',
                    'id', 'deploy_template_id', 'interface', 'step', 'args',
                    'priority']
        self.assertEqual(sorted(expected), sorted(col_names))

        self.assertIsInstance(deploy_template_steps.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(deploy_template_steps.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(deploy_template_steps.c.version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(deploy_template_steps.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(deploy_template_steps.c.deploy_template_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(deploy_template_steps.c.interface.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(deploy_template_steps.c.step.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(deploy_template_steps.c.args.type,
                              sqlalchemy.types.Text)
        self.assertIsInstance(deploy_template_steps.c.priority.type,
                              sqlalchemy.types.Integer)

        with engine.begin() as connection:
            # Insert a deploy template.
            uuid = uuidutils.generate_uuid()
            name = 'CUSTOM_DT1'
            template = {'name': name, 'uuid': uuid}
            insert_dpt = deploy_templates.insert().values(template)
            connection.execute(insert_dpt)
            # Query by UUID.
            dpt_uuid_stmt = sqlalchemy.select(
                models.DeployTemplate.id,
                models.DeployTemplate.name,
            ).where(
                models.DeployTemplate.uuid == uuid
            )
            result = connection.execute(dpt_uuid_stmt).first()
            template_id = result.id
            self.assertEqual(name, result.name)
            # Query by name.
            dpt_name_stmt = sqlalchemy.select(
                models.DeployTemplate.id
            ).where(
                models.DeployTemplate.name == name
            )
            result = connection.execute(dpt_name_stmt).first()
            self.assertEqual(template_id, result.id)
            # Query by ID.
            dpt_id_stmt = sqlalchemy.select(
                models.DeployTemplate.uuid,
                models.DeployTemplate.name
            ).where(
                models.DeployTemplate.id == template_id
            )
            result = connection.execute(dpt_id_stmt).first()
            self.assertEqual(uuid, result.uuid)
            self.assertEqual(name, result.name)
            savepoint_uuid = connection.begin_nested()
            # UUID is unique.
            template = {'name': 'CUSTOM_DT2', 'uuid': uuid}
            self.assertRaises(db_exc.DBDuplicateEntry, connection.execute,
                              deploy_templates.insert(), template)
            savepoint_uuid.rollback()
            savepoint_uuid.close()
            # Name is unique.
            savepoint_name = connection.begin_nested()
            template = {'name': name, 'uuid': uuidutils.generate_uuid()}
            self.assertRaises(db_exc.DBDuplicateEntry, connection.execute,
                              deploy_templates.insert(), template)
            savepoint_name.rollback()
            savepoint_name.close()

            # Insert a deploy template step.
            interface = 'raid'
            step_name = 'create_configuration'
            # The line below is JSON.
            args = '{"logical_disks": []}'
            priority = 10
            step = {'deploy_template_id': template_id, 'interface': interface,
                    'step': step_name, 'args': args, 'priority': priority}
            insert_dpts = deploy_template_steps.insert().values(step)
            connection.execute(insert_dpts)
            # Query by deploy template ID.
            query_id_stmt = sqlalchemy.select(
                models.DeployTemplateStep.deploy_template_id,
                models.DeployTemplateStep.interface,
                models.DeployTemplateStep.step,
                models.DeployTemplateStep.args,
                models.DeployTemplateStep.priority,
            ).where(
                models.DeployTemplateStep.deploy_template_id == template_id
            )
            result = connection.execute(query_id_stmt).first()
            self.assertEqual(template_id, result.deploy_template_id)
            self.assertEqual(interface, result.interface)
            self.assertEqual(step_name, result.step)
            if isinstance(result.args, dict):
                # Postgres testing results in a dict being returned
                # at this level which if you str() it, you get a dict,
                # so comparing string to string fails.
                result_args = json.dumps(result.args)
            else:
                # Mysql/MariaDB appears to be actually hand us
                # a string back so we should be able to compare it.
                result_args = result.args
            self.assertEqual(args, result_args)
            self.assertEqual(priority, result.priority)
            # Insert another step for the same template.
            insert_step = deploy_template_steps.insert().values(step)
            connection.execute(insert_step)

    def _check_1e15e7122cc9(self, engine, data):
        # Deploy template 'extra' field.
        deploy_templates = db_utils.get_table(engine, 'deploy_templates')
        col_names = [column.name for column in deploy_templates.c]
        expected = ['created_at', 'updated_at', 'version',
                    'id', 'uuid', 'name', 'extra']
        self.assertEqual(sorted(expected), sorted(col_names))
        self.assertIsInstance(deploy_templates.c.extra.type,
                              sqlalchemy.types.TEXT)

    def _check_dada631878c4(self, engine, data):
        # Runbooks.
        runbooks = db_utils.get_table(engine, 'runbooks')
        col_names = [column.name for column in runbooks.c]
        expected = ['created_at', 'updated_at', 'version',
                    'id', 'uuid', 'name']
        self.assertEqual(sorted(expected), sorted(col_names))
        self.assertIsInstance(runbooks.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(runbooks.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(runbooks.c.version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(runbooks.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(runbooks.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(runbooks.c.name.type,
                              sqlalchemy.types.String)

        # Runbook steps.
        runbook_steps = db_utils.get_table(engine, 'runbook_steps')
        col_names = [column.name for column in runbook_steps.c]
        expected = ['created_at', 'updated_at', 'version', 'id',
                    'runbook_id', 'interface', 'step', 'args',
                    'order']
        self.assertEqual(sorted(expected), sorted(col_names))

        self.assertIsInstance(runbook_steps.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(runbook_steps.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(runbook_steps.c.version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(runbook_steps.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(runbook_steps.c.runbook_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(runbook_steps.c.interface.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(runbook_steps.c.step.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(runbook_steps.c.args.type,
                              sqlalchemy.types.Text)
        self.assertIsInstance(runbook_steps.c.order.type,
                              sqlalchemy.types.Integer)

        with engine.begin() as connection:
            # Insert a Runbook.
            uuid = uuidutils.generate_uuid()
            name = 'CUSTOM_DT1'
            runbook = {'name': name, 'uuid': uuid}
            insert_dpt = runbooks.insert().values(runbook)
            connection.execute(insert_dpt)
            # Query by UUID.
            dpt_uuid_stmt = sqlalchemy.select(
                models.Runbook.id,
                models.Runbook.name,
            ).where(
                models.Runbook.uuid == uuid
            )
            result = connection.execute(dpt_uuid_stmt).first()
            runbook_id = result.id
            self.assertEqual(name, result.name)
            # Query by name.
            dpt_name_stmt = sqlalchemy.select(
                models.Runbook.id
            ).where(
                models.Runbook.name == name
            )
            result = connection.execute(dpt_name_stmt).first()
            self.assertEqual(runbook_id, result.id)
            # Query by ID.
            dpt_id_stmt = sqlalchemy.select(
                models.Runbook.uuid,
                models.Runbook.name
            ).where(
                models.Runbook.id == runbook_id
            )
            result = connection.execute(dpt_id_stmt).first()
            self.assertEqual(uuid, result.uuid)
            self.assertEqual(name, result.name)
            savepoint_uuid = connection.begin_nested()
            # UUID is unique.
            runbook = {'name': 'CUSTOM_DT2', 'uuid': uuid}
            self.assertRaises(db_exc.DBDuplicateEntry, connection.execute,
                              runbooks.insert(), runbook)
            savepoint_uuid.rollback()
            savepoint_uuid.close()
            # Name is unique.
            savepoint_name = connection.begin_nested()
            runbook = {'name': name, 'uuid': uuidutils.generate_uuid()}
            self.assertRaises(db_exc.DBDuplicateEntry, connection.execute,
                              runbooks.insert(), runbook)
            savepoint_name.rollback()
            savepoint_name.close()

            # Insert a Runbook step.
            interface = 'raid'
            step_name = 'create_configuration'
            # The line below is JSON.
            args = '{"logical_disks": []}'
            order = 1
            step = {'runbook_id': runbook_id, 'interface': interface,
                    'step': step_name, 'args': args, 'order': order}
            insert_dpts = runbook_steps.insert().values(step)
            connection.execute(insert_dpts)
            # Query by Runbook ID.
            query_id_stmt = sqlalchemy.select(
                models.RunbookStep.runbook_id,
                models.RunbookStep.interface,
                models.RunbookStep.step,
                models.RunbookStep.args,
                models.RunbookStep.order,
            ).where(
                models.RunbookStep.runbook_id == runbook_id
            )
            result = connection.execute(query_id_stmt).first()
            self.assertEqual(runbook_id, result.runbook_id)
            self.assertEqual(interface, result.interface)
            self.assertEqual(step_name, result.step)
            if isinstance(result.args, dict):
                # Postgres testing results in a dict being returned
                # at this level which if you str() it, you get a dict,
                # so comparing string to string fails.
                result_args = json.dumps(result.args)
            else:
                # Mysql/MariaDB appears to be actually hand us
                # a string back so we should be able to compare it.
                result_args = result.args
            self.assertEqual(args, result_args)
            self.assertEqual(order, result.order)
            # Insert another step for the same runbook.
            insert_step = runbook_steps.insert().values(step)
            connection.execute(insert_step)

    def _check_245c3e54b247(self, engine, data):
        # Runbook 'extra' field.
        runbooks = db_utils.get_table(engine, 'runbooks')
        col_names = [column.name for column in runbooks.c]
        expected = ['created_at', 'updated_at', 'version',
                    'id', 'uuid', 'name', 'extra']
        self.assertEqual(sorted(expected), sorted(col_names))
        self.assertIsInstance(runbooks.c.extra.type,
                              sqlalchemy.types.TEXT)

    def _check_ce6c4b3cf5a2(self, engine, data):
        allocations = db_utils.get_table(engine, 'allocations')
        col_names = [column.name for column in allocations.c]
        self.assertIn('owner', col_names)

    def _check_cf1a80fdb352(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('network_data', col_names)
        self.assertIsInstance(
            nodes.c.network_data.type, sqlalchemy.types.String)

    def _check_c1846a214450(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('boot_mode', col_names)
        self.assertIn('secure_boot', col_names)
        self.assertIsInstance(nodes.c.boot_mode.type,
                              sqlalchemy.types.String)
        # in some backends bool type is integer
        self.assertIsInstance(nodes.c.secure_boot.type,
                              (sqlalchemy.types.Boolean,
                               sqlalchemy.types.Integer))

    def _pre_upgrade_cd2c80feb331(self, engine):
        data = {
            'node_uuid': uuidutils.generate_uuid(),
        }

        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            insert_node = nodes.insert().values({'uuid': data['node_uuid']})
            connection.execute(insert_node)

        return data

    def _check_cd2c80feb331(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('retired', col_names)
        self.assertIn('retired_reason', col_names)

        with engine.begin() as connection:
            node_stmt = sqlalchemy.select(
                models.Node.retired,
                models.Node.retired_reason,
            ).where(
                models.Node.uuid == data['node_uuid']
            )
            node = connection.execute(node_stmt).first()
            self.assertFalse(node.retired)
            self.assertIsNone(node.retired_reason)

    def _check_b2ad35726bb0(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('lessee', col_names)

    def _check_c0455649680c(self, engine, data):
        ports = db_utils.get_table(engine, 'ports')
        col_names = [column.name for column in ports.c]
        self.assertIn('name', col_names)

    def _check_9ef41f07cb58(self, engine, data):
        node_history = db_utils.get_table(engine, 'node_history')
        col_names = [column.name for column in node_history.c]

        expected_names = ['version', 'created_at', 'updated_at', 'id', 'uuid',
                          'conductor', 'event_type', 'severity', 'event',
                          'user', 'node_id']
        self.assertEqual(sorted(expected_names), sorted(col_names))

        self.assertIsInstance(node_history.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(node_history.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(node_history.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(node_history.c.uuid.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(node_history.c.conductor.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(node_history.c.event_type.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(node_history.c.severity.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(node_history.c.event.type,
                              sqlalchemy.types.TEXT)
        self.assertIsInstance(node_history.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(node_history.c.user.type,
                              sqlalchemy.types.String)

    def _check_0ac0f39bc5aa(self, engine, data):
        node_inventory = db_utils.get_table(engine, 'node_inventory')
        col_names = [column.name for column in node_inventory.c]

        expected_names = ['version', 'created_at', 'updated_at', 'id',
                          'node_id', 'inventory_data', 'plugin_data']
        self.assertEqual(sorted(expected_names), sorted(col_names))

        self.assertIsInstance(node_inventory.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(node_inventory.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(node_inventory.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(node_inventory.c.node_id.type,
                              sqlalchemy.types.Integer)

    def _check_4dbec778866e(self, engine, data):
        nodes = db_utils.get_table(engine, 'nodes')
        self.assertIsInstance(nodes.c.shard.type, sqlalchemy.types.String)

    def _pre_upgrade_163040c5513f(self, engine):
        # Create a node to which firmware information can be added.
        data = {'uuid': uuidutils.generate_uuid()}
        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)
            node_stmt = sqlalchemy.select(
                models.Node.id
            ).where(models.Node.uuid == data['uuid'])
            node = connection.execute(node_stmt).first()
            # WARNING: Always copy, never directly return a db object or
            # piece of a db object. It is a sqlalchemy thing.
            data['id'] = int(node.id)

        return data

    def _check_163040c5513f(self, engine, data):
        fw_information = db_utils.get_table(engine, 'firmware_information')
        col_names = [column.name for column in fw_information.c]
        expected_names = ['created_at', 'updated_at', 'id', 'node_id',
                          'component', 'initial_version', 'current_version',
                          'last_version_flashed', 'version']
        self.assertEqual(sorted(expected_names), sorted(col_names))
        self.assertIsInstance(fw_information.c.created_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(fw_information.c.updated_at.type,
                              sqlalchemy.types.DateTime)
        self.assertIsInstance(fw_information.c.id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(fw_information.c.node_id.type,
                              sqlalchemy.types.Integer)
        self.assertIsInstance(fw_information.c.component.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(fw_information.c.initial_version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(fw_information.c.current_version.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(fw_information.c.last_version_flashed.type,
                              sqlalchemy.types.String)
        self.assertIsInstance(fw_information.c.version.type,
                              sqlalchemy.types.String)

        nodes = db_utils.get_table(engine, 'nodes')
        col_names = [column.name for column in nodes.c]
        self.assertIn('firmware_interface', col_names)
        self.assertIsInstance(nodes.c.firmware_interface.type,
                              sqlalchemy.types.String)

        with engine.begin() as connection:
            fw_data = {'node_id': data['id'],
                       'component': 'bmc',
                       'initial_version': 'v1.0.0'}
            insert_fw_cmp = fw_information.insert().values(fw_data)
            connection.execute(insert_fw_cmp)
            fw_cmp_stmt = sqlalchemy.select(
                models.FirmwareComponent.initial_version
            ).where(
                models.FirmwareComponent.node_id == data['id'],
                models.FirmwareComponent.component == fw_data['component']
            )
            fw_component = connection.execute(fw_cmp_stmt).first()
            self.assertEqual('v1.0.0', fw_component[0])
            del_stmt = (
                sqlalchemy.delete(
                    models.FirmwareComponent
                ).where(models.FirmwareComponent.node_id == data['id'])
            )
            connection.execute(del_stmt)

    def test_upgrade_and_version(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('head')
            self.assertIsNotNone(self.migration_api.version())

    def test_create_schema_and_version(self):
        with patch_with_engine(self.engine):
            self.migration_api.create_schema()
            self.assertIsNotNone(self.migration_api.version())

    def test_upgrade_and_create_schema(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('31baaf680d2b')
            self.assertRaises(db_exc.DBMigrationError,
                              self.migration_api.create_schema)

    def test_upgrade_twice(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('31baaf680d2b')
            v1 = self.migration_api.version()
            self.migration_api.upgrade('head')
            v2 = self.migration_api.version()
            self.assertNotEqual(v1, v2)


class TestMigrationsMySQL(MigrationCheckersMixin,
                          WalkVersionsMixin,
                          test_fixtures.OpportunisticDBTestMixin,
                          test_base.BaseTestCase):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture

    def _pre_upgrade_e918ff30eb42(self, engine):

        nodes = db_utils.get_table(engine, 'nodes')

        # this should always fail pre-upgrade
        mediumtext = 'a' * (pow(2, 16) + 1)
        json_text = str({'key': mediumtext})
        uuid = uuidutils.generate_uuid()
        expected_to_fail_data = {'uuid': uuid, 'instance_info': json_text}
        # this should always work pre-upgrade
        text = 'a' * (pow(2, 16) - 13)
        # The field needs to contain JSON for the decoder to work against
        # the field.
        json_text = str({'key': text})
        uuid2 = uuidutils.generate_uuid()
        valid_pre_upgrade_data = {'uuid': uuid2, 'instance_info': json_text}
        with engine.begin() as connection:
            self.assertRaises(db_exc.DBError, connection.execute,
                              nodes.insert(), expected_to_fail_data)

            insert_node = nodes.insert().values(valid_pre_upgrade_data)
            connection.execute(insert_node)

        return valid_pre_upgrade_data

    def _check_e918ff30eb42(self, engine, data):

        nodes = db_utils.get_table(engine, 'nodes')
        with engine.begin() as connection:
            # check that the data for the successful pre-upgrade
            # entry didn't change
            # NOTE(TheJulia): Directly select the field to bypass
            # field decoding
            i_info = connection.execute(
                sqlalchemy.text(
                    "SELECT instance_info from nodes WHERE uuid = "
                    "'%s'" % data['uuid'])).one()
            self.assertIsNotNone(i_info[0])
            self.assertEqual(data['instance_info'], i_info[0])

            # now this should pass post-upgrade
            test = 'b' * (pow(2, 16) + 1)
            test_text = str({'a': test})
            uuid = uuidutils.generate_uuid()
            data = {'uuid': uuid, 'instance_info': test_text}
            insert_node = nodes.insert().values(data)
            connection.execute(insert_node)

            # Re-uses the same query to fetch current results
            i_info = connection.execute(
                sqlalchemy.text(
                    "SELECT instance_info from nodes WHERE uuid = "
                    "'%s'" % data['uuid'])).one()
            self.assertEqual(test_text, i_info[0])

    def _check_0ac0f39bc5aa(self, engine, data):
        node_inventory = db_utils.get_table(engine, 'node_inventory')
        self.assertIsInstance(node_inventory.c.inventory_data.type,
                              sqlalchemy.dialects.mysql.LONGTEXT)
        self.assertIsInstance(node_inventory.c.plugin_data.type,
                              sqlalchemy.dialects.mysql.LONGTEXT)


class TestMigrationsPostgreSQL(MigrationCheckersMixin,
                               WalkVersionsMixin,
                               test_fixtures.OpportunisticDBTestMixin,
                               test_base.BaseTestCase):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture

    def _check_0ac0f39bc5aa(self, engine, data):
        node_inventory = db_utils.get_table(engine, 'node_inventory')
        self.assertIsInstance(node_inventory.c.inventory_data.type,
                              sqlalchemy.types.Text)
        self.assertIsInstance(node_inventory.c.plugin_data.type,
                              sqlalchemy.types.Text)


class ModelsMigrationSyncMixin(object):

    def setUp(self):
        super(ModelsMigrationSyncMixin, self).setUp()
        self.engine = enginefacade.writer.get_engine()
        self.useFixture(fixtures.Timeout(int(MIGRATIONS_TIMEOUT),
                                         gentle=False))

    def get_metadata(self):
        return models.Base.metadata

    def get_engine(self):
        return self.engine

    def db_sync(self, engine):
        with patch_with_engine(engine):
            migration.upgrade('head')


class ModelsMigrationsSyncMysql(ModelsMigrationSyncMixin,
                                test_migrations.ModelsMigrationsSync,
                                test_fixtures.OpportunisticDBTestMixin,
                                test_base.BaseTestCase):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils


INFO_DICT = db_utils.get_test_drac_info()


class BaseDracTest(db_base.DbTestCase):
    def setUp(self):
        super(BaseDracTest, self).setUp()
        self.config(enabled_hardware_types=['idrac', 'fake-hardware'],
                    enabled_boot_interfaces=[
                        'idrac-redfish-virtual-media', 'fake'],
                    enabled_power_interfaces=['idrac-redfish', 'fake'],
                    enabled_management_interfaces=['idrac-redfish', 'fake'],
                    enabled_inspect_interfaces=[
                        'idrac-redfish', 'fake', 'no-inspect'],
                    enabled_vendor_interfaces=[
                        'idrac-redfish', 'fake', 'no-vendor'],
                    enabled_raid_interfaces=['idrac-redfish', 'fake',
                                             'no-raid'],
                    enabled_bios_interfaces=['idrac-redfish', 'no-bios'])


class DictToObj(object):
    """Returns a dictionary into a class"""
    def __init__(self, dictionary):
        for key in dictionary:
            setattr(self, key, dictionary[key])


def dict_to_namedtuple(name='GenericNamedTuple', values=None,
                       tuple_class=None):
    """Converts a dict to a collections.namedtuple"""

    if values is None:
        values = {}

    if tuple_class is None:
        tuple_class = collections.namedtuple(name, list(values))
    else:
        # Support different versions of the driver as fields change.
        values = {field: values.get(field) for field in tuple_class._fields}

    return tuple_class(**values)


def dict_of_object(data):
    """Create a dictionary object"""

    for k, v in data.items():
        if isinstance(v, dict):
            dict_obj = DictToObj(v)
            data[k] = dict_obj
    return data


def create_raid_setting(raid_settings_dict):
    """Returns the raid configuration tuple object"""
    return dict_to_namedtuple(values=raid_settings_dict)



# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from hacking import core

# N323: Found use of _() without explicit import of _!


UNDERSCORE_IMPORT_FILES = []


string_translation = re.compile(r"[^_]*_\(\s*('|\")")
translated_log = re.compile(
    r"(.)*LOG\.(audit|error|info|warn|warning|critical|exception)"
    r"\(\s*_\(\s*('|\")")
underscore_import_check = re.compile(r"(.)*import _(.)*")
# We need this for cases where they have created their own _ function.
custom_underscore_check = re.compile(r"(.)*_\s*=\s*(.)*")


@core.flake8ext
def check_explicit_underscore_import(logical_line, filename):
    """Check for explicit import of the _ function

    We need to ensure that any files that are using the _() function
    to translate logs are explicitly importing the _ function.  We
    can't trust unit test to catch whether the import has been
    added so we need to check for it here.

    """

    # Build a list of the files that have _ imported.  No further
    # checking needed once it is found.
    if filename in UNDERSCORE_IMPORT_FILES:
        pass
    elif (underscore_import_check.match(logical_line)
          or custom_underscore_check.match(logical_line)):
        UNDERSCORE_IMPORT_FILES.append(filename)
    elif (translated_log.match(logical_line)
          or string_translation.match(logical_line)):
        yield (0, "N323: Found use of _() without explicit import of _!")



# coding=utf-8
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import types
from unittest import mock

from oslo_utils import timeutils

from ironic.common import exception
from ironic import objects
from ironic.objects import base
from ironic.objects import fields
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils


class TestConductorObject(db_base.DbTestCase):

    def setUp(self):
        super(TestConductorObject, self).setUp()
        self.fake_conductor = (
            db_utils.get_test_conductor(updated_at=timeutils.utcnow()))

    def test_load(self):
        host = self.fake_conductor['hostname']
        with mock.patch.object(self.dbapi, 'get_conductor',
                               autospec=True) as mock_get_cdr:
            mock_get_cdr.return_value = self.fake_conductor
            objects.Conductor.get_by_hostname(self.context, host)
            mock_get_cdr.assert_called_once_with(host, online=True)

    def test_list(self):
        conductor1 = db_utils.get_test_conductor(hostname='cond1')
        conductor2 = db_utils.get_test_conductor(hostname='cond2')
        with mock.patch.object(self.dbapi, 'get_conductor_list',
                               autospec=True) as mock_cond_list:
            mock_cond_list.return_value = [conductor1, conductor2]
            conductors = objects.Conductor.list(self.context)
            self.assertEqual(2, len(conductors))
            self.assertIsInstance(conductors[0], objects.Conductor)
            self.assertIsInstance(conductors[1], objects.Conductor)
            self.assertEqual(conductors[0].hostname, 'cond1')
            self.assertEqual(conductors[1].hostname, 'cond2')

    def test_save(self):
        host = self.fake_conductor['hostname']
        with mock.patch.object(self.dbapi, 'get_conductor',
                               autospec=True) as mock_get_cdr:
            mock_get_cdr.return_value = self.fake_conductor
            c = objects.Conductor.get_by_hostname(self.context, host)
            c.hostname = 'another-hostname'
            self.assertRaises(NotImplementedError,
                              c.save, self.context)
            mock_get_cdr.assert_called_once_with(host, online=True)

    def test_touch(self):
        host = self.fake_conductor['hostname']
        with mock.patch.object(self.dbapi, 'get_conductor',
                               autospec=True) as mock_get_cdr:
            with mock.patch.object(self.dbapi, 'touch_conductor',
                                   autospec=True) as mock_touch_cdr:
                mock_get_cdr.return_value = self.fake_conductor
                c = objects.Conductor.get_by_hostname(self.context, host)
                c.touch(self.context)
                mock_get_cdr.assert_called_once_with(host, online=True)
                mock_touch_cdr.assert_called_once_with(host, online=True)

    def test_refresh(self):
        host = self.fake_conductor['hostname']
        t0 = self.fake_conductor['updated_at']
        t1 = t0 + datetime.timedelta(seconds=10)
        returns = [dict(self.fake_conductor, updated_at=t0),
                   dict(self.fake_conductor, updated_at=t1)]
        expected = [mock.call(host, online=True),
                    mock.call(host, online=True)]
        with mock.patch.object(self.dbapi, 'get_conductor',
                               side_effect=returns,
                               autospec=True) as mock_get_cdr:
            c = objects.Conductor.get_by_hostname(self.context, host)
            # ensure timestamps have tzinfo
            datetime_field = fields.DateTimeField()
            self.assertEqual(
                datetime_field.coerce(datetime_field, 'updated_at', t0),
                c.updated_at)
            c.refresh()
            self.assertEqual(
                datetime_field.coerce(datetime_field, 'updated_at', t1),
                c.updated_at)
            self.assertEqual(expected, mock_get_cdr.call_args_list)
            self.assertEqual(self.context, c._context)

    @mock.patch.object(base.IronicObject, 'get_target_version',
                       spec_set=types.FunctionType)
    def _test_register(self, mock_target_version, update_existing=False,
                       conductor_group=''):
        mock_target_version.return_value = '1.5'
        host = self.fake_conductor['hostname']
        drivers = self.fake_conductor['drivers']
        with mock.patch.object(self.dbapi, 'register_conductor',
                               autospec=True) as mock_register_cdr:
            mock_register_cdr.return_value = self.fake_conductor
            c = objects.Conductor.register(self.context, host, drivers,
                                           conductor_group,
                                           update_existing=update_existing)

            self.assertIsInstance(c, objects.Conductor)
            mock_register_cdr.assert_called_once_with(
                {'drivers': drivers,
                 'hostname': host,
                 'conductor_group': conductor_group.lower(),
                 'version': '1.5'},
                update_existing=update_existing)

    def test_register(self):
        self._test_register()

    def test_register_update_existing_true(self):
        self._test_register(update_existing=True)

    def test_register_into_group(self):
        self._test_register(conductor_group='dc1')

    def test_register_into_group_uppercased(self):
        self._test_register(conductor_group='DC1')

    def test_register_into_group_with_update(self):
        self._test_register(conductor_group='dc1', update_existing=True)

    @mock.patch.object(base.IronicObject, 'get_target_version',
                       spec_set=types.FunctionType)
    def test_register_with_invalid_group(self, mock_target_version):
        mock_target_version.return_value = '1.5'
        host = self.fake_conductor['hostname']
        drivers = self.fake_conductor['drivers']
        self.assertRaises(exception.InvalidConductorGroup,
                          objects.Conductor.register,
                          self.context, host, drivers, 'invalid:group')

    @mock.patch.object(objects.Conductor, 'unregister_all_hardware_interfaces',
                       autospec=True)
    def test_unregister(self, mock_unreg_ifaces):
        host = self.fake_conductor['hostname']
        with mock.patch.object(self.dbapi, 'get_conductor',
                               autospec=True) as mock_get_cdr:
            with mock.patch.object(self.dbapi, 'unregister_conductor',
                                   autospec=True) as mock_unregister_cdr:
                mock_get_cdr.return_value = self.fake_conductor
                c = objects.Conductor.get_by_hostname(self.context, host)
                c.unregister()
                mock_unregister_cdr.assert_called_once_with(host)
                mock_unreg_ifaces.assert_called_once_with(mock.ANY)

    def test_register_hardware_interfaces(self):
        host = self.fake_conductor['hostname']
        self.config(default_deploy_interface='ansible')
        arg = [{"hardware_type": "hardware-type", "interface_type": "deploy",
                "interface_name": "ansible", "default": True},
               {"hardware_type": "hardware-type", "interface_type": "deploy",
                "interface_name": "direct", "default": False}]
        with mock.patch.object(self.dbapi, 'get_conductor',
                               autospec=True) as mock_get_cdr:
            with mock.patch.object(self.dbapi,
                                   'register_conductor_hardware_interfaces',
                                   autospec=True) as mock_register:
                mock_get_cdr.return_value = self.fake_conductor
                c = objects.Conductor.get_by_hostname(self.context, host)
                c.register_hardware_interfaces(arg)
                mock_register.assert_called_once_with(c.id, arg)

    def test_unregister_all_hardware_interfaces(self):
        host = self.fake_conductor['hostname']
        with mock.patch.object(self.dbapi, 'get_conductor',
                               autospec=True) as mock_get_cdr:
            with mock.patch.object(self.dbapi,
                                   'unregister_conductor_hardware_interfaces',
                                   autospec=True) as mock_unregister:
                mock_get_cdr.return_value = self.fake_conductor
                c = objects.Conductor.get_by_hostname(self.context, host)
                c.unregister_all_hardware_interfaces()
                mock_unregister.assert_called_once_with(c.id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
import yaml

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.inspection_rules import actions
from ironic.common.inspection_rules import operators
from ironic.common.inspection_rules import utils
from ironic.common.inspection_rules import validation
from ironic.conf import CONF
from ironic import objects


LOG = log.getLogger(__name__)
SENSITIVE_FIELDS = ['password', 'auth_token', 'bmc_password']


def get_built_in_rules():
    """Load built-in inspection rules."""
    built_in_rules = []
    built_in_rules_dir = CONF.inspection_rules.built_in_rules

    if not built_in_rules_dir:
        return built_in_rules

    try:
        with open(built_in_rules_dir, 'r') as f:
            rules_data = yaml.safe_load(f)

        for rule_data in rules_data:
            try:
                rule = {
                    'uuid': rule_data.get('uuid'),
                    'priority': rule_data.get('priority', 0),
                    'description': rule_data.get('description'),
                    'scope': rule_data.get('scope'),
                    'sensitive': rule_data.get('sensitive', False),
                    'phase': rule_data.get('phase', 'main'),
                    'actions': rule_data.get('actions', []),
                    'conditions': rule_data.get('conditions', []),
                    'built_in': True
                }
                validation.validate_rule(rule)
                built_in_rules.append(rule)
            except Exception as e:
                LOG.error(_("Error parsing built-in rule: %s"), e)
                raise
    except FileNotFoundError:
        LOG.error(_("Built-in rules file not found: %s"),
                  built_in_rules_dir)
        raise
    except yaml.YAMLError as e:
        LOG.error(_("Error parsing YAML in built-in rules file %s: %s"),
                  built_in_rules_dir, e)
        raise
    except Exception as e:
        LOG.error(_("Error loading built-in rules from %s: %s"),
                  built_in_rules_dir, e)
        raise

    return built_in_rules


def check_conditions(task, rule, inventory, plugin_data):
    try:
        if not rule.get('conditions', None):
            return True

        for condition in rule['conditions']:
            op, invtd = utils.parse_inverted_operator(
                condition['op'])

            if op not in operators.OPERATORS:
                supported_ops = ', '.join(operators.OPERATORS.keys())
                msg = (_("Unsupported operator: '%(op)s'. Supported "
                         "operators are: %(supported_ops)s.") % {
                             'op': op, 'supported_ops': supported_ops})
                raise ValueError(msg)

            plugin = operators.get_operator(op)
            if condition.get('loop', []):
                result = plugin().check_with_loop(task, condition, inventory,
                                                  plugin_data)
            else:
                result = plugin().check_condition(task, condition, inventory,
                                                  plugin_data)
            if not result:
                LOG.debug("Skipping rule %(rule)s on node %(node)s: "
                          "condition check '%(op)s': '%(args)s' failed ",
                          {'rule': rule['uuid'], 'node': task.node.uuid,
                           'op': condition['op'], 'args': condition['args']})
                return False
        return True

    except Exception as err:
        LOG.error("Error checking condition on node %(node)s: %(err)s.",
                  {'node': task.node.uuid, 'err': err})
        raise


def apply_actions(task, rule, inventory, plugin_data):

    for action in rule['actions']:
        try:
            op = action['op']
            if op not in actions.ACTIONS:
                supported_ops = ', '.join(actions.ACTIONS.keys())
                msg = (_("Unsupported action: '%(op)s'. Supported actions "
                         "are: %(supported_ops)s.") % {
                             'op': op, 'supported_ops': supported_ops})
                raise ValueError(msg)

            plugin = actions.get_action(op)
            if action.get('loop', []):
                plugin().execute_with_loop(task, action, inventory,
                                           plugin_data)
            else:
                plugin().execute_action(task, action, inventory,
                                        plugin_data)
        except exception.IronicException as err:
            LOG.error("Error applying action on node %(node)s: %(err)s.",
                      {'node': task.node.uuid, 'err': err})
            raise
        except Exception as err:
            LOG.exception("Unexpected error applying action on node "
                          "%(node)s: %(err)s.", {'node': task.node.uuid,
                                                 'err': err})
            raise


def apply_rules(task, inventory, plugin_data, inspection_phase):
    """Apply inspection rules to a node."""
    node = task.node

    all_rules = objects.InspectionRule.list(
        context=task.context,
        filters={'phase': inspection_phase})

    built_in_rules = get_built_in_rules()
    rules = all_rules + built_in_rules

    if not rules:
        LOG.debug("No inspection rules to apply for phase "
                  "'%(phase)s on node: %(node)s'", {
                      'phase': inspection_phase,
                      'node': node.uuid})
        return

    rules.sort(key=lambda rule: rule.get('priority', 0), reverse=True)
    LOG.debug("Applying %(count)d inspection rules to node %(node)s",
              {'count': len(rules), 'node': node.uuid})

    mask_secrets = CONF.inspection_rules.mask_secrets
    for rule in rules:
        try:

            should_mask = False
            is_sensitive_rule = rule.get('sensitive', False)

            if (mask_secrets == 'always'
                or mask_secrets == 'sensitive' and not is_sensitive_rule):
                should_mask = True

            masked_inventory = utils.ShallowMaskDict(
                inventory, sensitive_fields=SENSITIVE_FIELDS,
                mask_enabled=should_mask)

            masked_plugin_data = utils.ShallowMaskDict(
                plugin_data, sensitive_fields=SENSITIVE_FIELDS,
                mask_enabled=should_mask)

            if not check_conditions(task, rule, masked_inventory,
                                    masked_plugin_data):
                continue

            LOG.info("Applying actions for rule %(rule)s to node %(node)s",
                     {'rule': rule['uuid'], 'node': node.uuid})

            apply_actions(task, rule, masked_inventory, masked_plugin_data)

        except exception.HardwareInspectionFailure:
            raise
        except exception.IronicException as e:
            LOG.error(_("Error applying rule %(rule)s to node "
                        "%(node)s: %(error)s"), {'rule': rule['uuid'],
                                                 'node': node.uuid,
                                                 'error': e})
            raise
        except Exception as e:
            msg = ("Failed to apply rule %(rule)s to node %(node)s: "
                   "%(error)s" % {'rule': rule['uuid'], 'node': node.uuid,
                                  'error': e})

            LOG.exception(msg)

            raise exception.IronicException(msg)

    LOG.info("Finished applying inspection rules to node %s", node.uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import types
from unittest import mock


from ironic.common import exception
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils


class TestFirmwareComponentObject(db_base.DbTestCase,
                                  obj_utils.SchemasTestMixIn):

    def setUp(self):
        super(TestFirmwareComponentObject, self).setUp()
        self.firmware_component_dict = db_utils.get_test_firmware_component()

    def test_get_firmware_component(self):
        with mock.patch.object(self.dbapi, 'get_firmware_component',
                               autospec=True) as mock_get_component:
            node_id = self.firmware_component_dict['node_id']
            name = self.firmware_component_dict['component']
            mock_get_component.return_value = self.firmware_component_dict

            fw_component = objects.FirmwareComponent.get(
                self.context, node_id, name)

            mock_get_component.assert_called_once_with(node_id, name)

            self.assertEqual(self.context, fw_component._context)
            self.assertEqual(self.firmware_component_dict['node_id'],
                             fw_component.node_id)
            self.assertEqual(self.firmware_component_dict['component'],
                             fw_component.component)
            self.assertEqual(self.firmware_component_dict['initial_version'],
                             fw_component.initial_version)
            self.assertEqual(self.firmware_component_dict['current_version'],
                             fw_component.current_version)
            self.assertEqual(
                self.firmware_component_dict['last_version_flashed'],
                fw_component.last_version_flashed)

    def test_get_firmware_component_does_not_exist(self):
        fw_cmp_name = "does not exist"
        with mock.patch.object(self.dbapi, 'get_firmware_component',
                               autospec=True) as mock_get_component:
            not_found = exception.FirmwareComponentNotFound(
                node=self.firmware_component_dict['node_id'],
                name=fw_cmp_name)
            mock_get_component.side_effect = not_found

            self.assertRaises(exception.FirmwareComponentNotFound,
                              objects.FirmwareComponent.get, self.context,
                              self.firmware_component_dict['node_id'],
                              fw_cmp_name)

    def test_get_firmware_component_node_does_not_exist(self):
        with mock.patch.object(self.dbapi, 'get_firmware_component',
                               autospec=True) as mock_get_component:
            mock_get_component.side_effect = exception.NodeNotFound(node=404)

            self.assertRaises(exception.NodeNotFound,
                              objects.FirmwareComponent.get, self.context,
                              self.firmware_component_dict['node_id'],
                              self.firmware_component_dict['component'])

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_firmware_component',
                               autospec=True) as mock_db_create:
            mock_db_create.return_value = self.firmware_component_dict
            new_fw_component = objects.FirmwareComponent(
                self.context, **self.firmware_component_dict)
            new_fw_component.create()

            mock_db_create.assert_called_once_with(
                self.firmware_component_dict)

    def test_save(self):
        node_id = self.firmware_component_dict['node_id']
        name = self.firmware_component_dict['component']
        test_time = datetime.datetime(2000, 1, 1, 0, 0)

        with mock.patch.object(self.dbapi, 'get_firmware_component',
                               autospec=True) as mock_get_fw:
            mock_get_fw.return_value = self.firmware_component_dict
            with mock.patch.object(self.dbapi, 'update_firmware_component',
                                   autospec=True) as mock_update_fw:

                mock_update_fw.return_value = (
                    db_utils.get_test_firmware_component(
                        current_version="v2.0.1",
                        last_version_flashed="v2.0.1",
                        updated_at=test_time)
                )

                fw_cmp = objects.FirmwareComponent.get(
                    self.context, node_id, name)
                fw_cmp.current_version = "v2.0.1"
                fw_cmp.last_version_flashed = "v2.0.1"
                fw_cmp.save()

                mock_get_fw.assert_called_once_with(node_id, name)

                mock_update_fw.assert_called_once_with(
                    node_id, fw_cmp.component,
                    {'current_version': 'v2.0.1',
                     'last_version_flashed': 'v2.0.1',
                     'version': objects.FirmwareComponent.VERSION})

                self.assertEqual(self.context, fw_cmp._context)
                self.assertEqual("v2.0.1", fw_cmp.current_version)
                self.assertEqual("v2.0.1", fw_cmp.last_version_flashed)
                res_updated_at = (fw_cmp.updated_at).replace(tzinfo=None)
                self.assertEqual(test_time, res_updated_at)

    @mock.patch(
        'ironic.objects.firmware.FirmwareComponentList.get_by_node_id',
        spec_set=types.FunctionType)
    def test_sync_firmware_components_create_and_update(self, mock_get):
        node = obj_utils.create_test_node(self.context)
        fw_obj = obj_utils.create_test_firmware_component(
            self.context, node_id=node.id)
        mock_get.return_value = [fw_obj]
        components = db_utils.get_test_firmware_component_list()
        components[0]['current_version'] = 'v2.0.0'
        components[0]['last_version_flashed'] = 'v2.0.0'
        create, update, unchanged = (
            objects.FirmwareComponentList.sync_firmware_components(
                self.context, node.id, components))
        self.assertEqual(create, components[1:])
        self.assertEqual(update, [components[0]])
        self.assertEqual(unchanged, [])

    @mock.patch(
        'ironic.objects.firmware.FirmwareComponentList.get_by_node_id',
        spec_set=types.FunctionType)
    def test_sync_firmware_components_nochange(self, mock_get):
        node = obj_utils.create_test_node(self.context)
        fw_obj_1 = obj_utils.create_test_firmware_component(
            self.context, node_id=node.id)
        fw_obj_2 = obj_utils.create_test_firmware_component(
            self.context, node_id=node.id, component='BIOS',
            initial_version='v1.5.0', current_version='v1.5.0')
        mock_get.return_value = [fw_obj_1, fw_obj_2]
        components = db_utils.get_test_firmware_component_list()

        create, update, unchanged = (
            objects.FirmwareComponentList.sync_firmware_components(
                self.context, node.id, components))
        self.assertEqual(create, [])
        self.assertEqual(update, [])
        self.assertEqual(unchanged, components)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for DRAC RAID interface
"""

from unittest import mock

from oslo_utils import importutils
import sushy
import tenacity

from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.drac import raid as drac_raid
from ironic.drivers.modules.drac import utils as drac_utils
from ironic.drivers.modules.redfish import raid as redfish_raid
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.tests.unit.drivers.modules.drac import utils as test_utils
from ironic.tests.unit.objects import utils as obj_utils

sushy_oem_idrac = importutils.try_import('sushy_oem_idrac')

INFO_DICT = test_utils.INFO_DICT


class DracCreateRaidConfigurationHelpersTestCase(test_utils.BaseDracTest):

    def setUp(self):
        super(DracCreateRaidConfigurationHelpersTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               driver='idrac',
                                               driver_info=INFO_DICT)

        self.physical_disk = {
            'id': 'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            'description': ('Disk 1 in Backplane 1 of '
                            'Integrated RAID Controller 1'),
            'controller': 'RAID.Integrated.1-1',
            'manufacturer': 'SEAGATE',
            'model': 'ST600MM0006',
            'media_type': 'hdd',
            'interface_type': 'sas',
            'size_mb': 571776,
            'free_size_mb': 571776,
            'serial_number': 'S0M3EY2Z',
            'firmware_version': 'LS0A',
            'status': 'ok',
            'raid_status': 'ready',
            'sas_address': '500056B37789ABE3',
            'device_protocol': None}

        self.physical_disks = []
        for i in range(8):
            disk = self.physical_disk.copy()
            disk['id'] = ('Disk.Bay.%s:Enclosure.Internal.0-1:'
                          'RAID.Integrated.1-1' % i)
            disk['serial_number'] = 'serial%s' % i

            self.physical_disks.append(disk)

        self.root_logical_disk = {
            'size_gb': 50,
            'raid_level': '1',
            'disk_type': 'hdd',
            'interface_type': 'sas',
            'volume_name': 'root_volume',
            'is_root_volume': True
        }
        self.nonroot_logical_disks = [
            {'size_gb': 100,
             'raid_level': '5',
             'disk_type': 'hdd',
             'interface_type': 'sas',
             'volume_name': 'data_volume1'},
            {'size_gb': 100,
             'raid_level': '5',
             'disk_type': 'hdd',
             'interface_type': 'sas',
             'volume_name': 'data_volume2'}
        ]

        self.logical_disks = (
            [self.root_logical_disk] + self.nonroot_logical_disks)
        self.target_raid_configuration = {'logical_disks': self.logical_disks}
        self.node.target_raid_config = self.target_raid_configuration
        self.node.save()


class DracRedfishRAIDTestCase(test_utils.BaseDracTest):

    def setUp(self):
        super(DracRedfishRAIDTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               driver='idrac',
                                               driver_info=INFO_DICT)
        self.raid = drac_raid.DracRedfishRAID()

    @mock.patch.object(drac_raid, '_wait_till_realtime_ready', autospec=True)
    @mock.patch.object(redfish_raid.RedfishRAID, 'create_configuration',
                       autospec=True)
    def test_create_configuration(self, mock_redfish_create, mock_wait):
        task = mock.Mock(node=self.node, context=self.context)

        self.raid.create_configuration(task)

        mock_wait.assert_called_once_with(task)
        mock_redfish_create.assert_called_once_with(
            self.raid, task, True, True, False)

    @mock.patch.object(drac_raid, '_wait_till_realtime_ready', autospec=True)
    @mock.patch.object(redfish_raid.RedfishRAID, 'delete_configuration',
                       autospec=True)
    def test_delete_configuration(self, mock_redfish_delete, mock_wait):
        task = mock.Mock(node=self.node, context=self.context)

        self.raid.delete_configuration(task)

        mock_wait.assert_called_once_with(task)
        mock_redfish_delete.assert_called_once_with(self.raid, task)

    @mock.patch.object(drac_raid, '_retry_till_realtime_ready', autospec=True)
    def test__wait_till_realtime_ready(self, mock_ready):
        self.node.set_driver_internal_info('cleaning_disable_ramdisk', True)
        task = mock.Mock(node=self.node, context=self.context)
        task.driver.power.get_power_state.return_value = states.POWER_OFF
        drac_raid._wait_till_realtime_ready(task)
        task.driver.power.set_power_state.assert_called_once_with(
            task, states.POWER_ON)
        mock_ready.assert_called_once_with(task)

    @mock.patch.object(drac_raid, 'LOG', autospec=True)
    @mock.patch.object(drac_raid, '_retry_till_realtime_ready', autospec=True)
    def test__wait_till_realtime_ready_retryerror(self, mock_ready, mock_log):
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.side_effect = tenacity.RetryError(3)
        drac_raid._wait_till_realtime_ready(task)
        mock_ready.assert_called_once_with(task)
        self.assertEqual(mock_log.debug.call_count, 1)

    @mock.patch.object(drac_raid, '_is_realtime_ready', autospec=True)
    def test__retry_till_realtime_ready_retry_exceeded(self, mock_ready):
        drac_raid._retry_till_realtime_ready.retry.sleep = mock.Mock()
        drac_raid._retry_till_realtime_ready.retry.stop =\
            tenacity.stop_after_attempt(3)
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.return_value = False

        self.assertRaises(
            tenacity.RetryError,
            drac_raid._retry_till_realtime_ready, task)

        self.assertEqual(3, mock_ready.call_count)

    @mock.patch.object(drac_raid, '_is_realtime_ready', autospec=True)
    def test__retry_till_realtime_ready_retry_fails(self, mock_ready):
        drac_raid._retry_till_realtime_ready.retry.sleep = mock.Mock()
        drac_raid._retry_till_realtime_ready.retry.stop =\
            tenacity.stop_after_attempt(3)
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.side_effect = [False, exception.RedfishError]

        self.assertRaises(
            exception.RedfishError,
            drac_raid._retry_till_realtime_ready, task)

        self.assertEqual(2, mock_ready.call_count)

    @mock.patch.object(drac_raid, '_is_realtime_ready', autospec=True)
    def test__retry_till_realtime_ready(self, mock_ready):
        drac_raid._retry_till_realtime_ready.retry.sleep = mock.Mock()
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.side_effect = [False, True]

        is_ready = drac_raid._retry_till_realtime_ready(task)

        self.assertTrue(is_ready)
        self.assertEqual(2, mock_ready.call_count)

    @mock.patch.object(drac_utils, 'LOG', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__is_realtime_ready(self, mock_get_system, mock_log):
        task = mock.Mock(node=self.node, context=self.context)
        fake_manager_oem1 = mock.Mock()
        fake_manager_oem1.lifecycle_service.is_realtime_ready.side_effect = (
            sushy.exceptions.SushyError)
        fake_manager1 = mock.Mock()
        fake_manager1.get_oem_extension.return_value = fake_manager_oem1
        fake_manager_oem2 = mock.Mock()
        fake_manager_oem2.lifecycle_service.is_realtime_ready.return_value = (
            True)
        fake_manager2 = mock.Mock()
        fake_manager2.get_oem_extension.return_value = fake_manager_oem2
        fake_system = mock.Mock(managers=[fake_manager1, fake_manager2])
        mock_get_system.return_value = fake_system

        is_ready = drac_raid._is_realtime_ready(task)

        self.assertTrue(is_ready)
        self.assertEqual(mock_log.debug.call_count, 1)

    def test_validate_correct_vendor(self):
        task = mock.Mock(node=self.node, context=self.context)
        self.node.properties['vendor'] = 'Dell Inc.'
        self.raid.validate(task)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_pre_create_configuration(self, mock_get_system):
        mock_task_mon1 = mock.Mock(check_is_processing=True)
        mock_task_mon2 = mock.Mock(check_is_processing=False)
        fake_oem_system = mock.Mock()
        fake_oem_system.change_physical_disk_state.return_value = [
            mock_task_mon1, mock_task_mon2]
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system

        mock_drive1 = mock.Mock(
            identity='Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1')
        mock_drive2 = mock.Mock(
            identity='Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            capacity_bytes=599550590976)  # mocked size in RAID mode
        mock_drive3 = mock.Mock(
            identity='Disk.Direct.0-0:AHCI.Slot.2-1')

        mock_controller1 = mock.Mock()
        mock_storage1 = mock.Mock(controllers=[mock_controller1],
                                  drives=[mock_drive1, mock_drive2],
                                  identity='RAID.Integrated.1-1')
        mock_storage1.controllers = mock.MagicMock()
        mock_storage1.controllers.get_members.return_value = [
            mock_controller1]

        mock_controller2 = mock.Mock()
        mock_storage2 = mock.Mock(controllers=[mock_controller2],
                                  drives=[mock_drive3],
                                  identity='AHCI.Slot.2-1')
        mock_storage2.controllers = mock.MagicMock()
        mock_storage2.controllers.get_members.return_value = [
            mock_controller2]

        fake_system.storage.get_members.return_value = [
            mock_storage1, mock_storage2]

        mock_get_system.return_value = fake_system
        task = mock.Mock(node=self.node, context=self.context)

        logical_disks_to_create = [{
            'raid_level': '0',
            'size_bytes': 600087461888,  # before RAID conversion
            'physical_disks': [
                'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'],
            'span_depth': 1,
            'span_length': 1.0,
            'controller': 'RAID.Integrated.1-1'}]

        result = self.raid.pre_create_configuration(
            task, logical_disks_to_create)

        self.assertEqual(
            [{'controller': 'RAID.Integrated.1-1',
              'physical_disks': [
                  'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'],
              'raid_level': '0',
              'size_bytes': 599550590976,  # recalculated after RAID conversion
              'span_depth': 1,
              'span_length': 1.0}], result)
        fake_oem_system.change_physical_disk_state.assert_called_once_with(
            sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_RAID,
            {mock_controller1: [mock_drive2]})
        mock_task_mon1.wait.assert_called_once_with(CONF.drac.raid_job_timeout)
        mock_task_mon2.wait.assert_not_called()

    def test__get_storage_controller_invalid_identity(self):
        fake_system = mock.Mock()

        mock_storage1 = mock.Mock(storage_controllers=[mock.Mock()],
                                  identity='RAID.Integrated.1-1')
        mock_storage2 = mock.Mock(storage_controllers=[mock.Mock()],
                                  identity='AHCI.Slot.2-1')

        fake_system.storage.get_members.return_value = [
            mock_storage1, mock_storage2]

        self.assertRaises(
            exception.IronicException,
            drac_raid.DracRedfishRAID._get_storage_controller,
            fake_system, 'NonExisting')

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    def test__change_physical_disk_state_attribute_error(self, mock_log):
        fake_oem_system = mock.Mock(spec=[])
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system

        result = drac_raid.DracRedfishRAID._change_physical_disk_state(
            fake_system, sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_RAID
        )

        self.assertEqual(False, result)
        mock_log.assert_called_once()

    @mock.patch.object(deploy_utils, 'reboot_to_finish_step',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode(
            self, mock_get_system, mock_reboot):
        mock_task_mon = mock.Mock(task_monitor_uri='/TaskService/1')
        mock_oem_controller = mock.Mock()
        mock_oem_controller.convert_to_raid.return_value = mock_task_mon
        mock_controller = mock.Mock()
        mock_controller.get_oem_extension.return_value = mock_oem_controller
        mock_controllers_col = mock.Mock()
        mock_controllers_col.get_members.return_value = [mock_controller]
        mock_storage = mock.Mock(controllers=mock_controllers_col)
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertEqual(
                ['/TaskService/1'],
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertEqual(mock_reboot.return_value, result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_no_conversion(
            self, mock_get_system):
        mock_oem_controller = mock.Mock()
        mock_oem_controller.convert_to_raid.return_value = None
        mock_controller = mock.Mock()
        mock_controller.get_oem_extension.return_value = mock_oem_controller
        mock_controllers_col = mock.Mock()
        mock_controllers_col.get_members.return_value = [mock_controller]
        mock_storage = mock.Mock(controllers=mock_controllers_col)
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_not_raid(
            self, mock_get_system, mock_log):
        mock_storage = mock.Mock(storage_controllers=None)
        mock_controllers = mock.PropertyMock(
            side_effect=sushy.exceptions.MissingAttributeError)
        type(mock_storage).controllers = mock_controllers
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)
            mock_log.assert_not_called()

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_old_idrac(
            self, mock_get_system, mock_log):
        mock_storage = mock.Mock(storage_controllers=mock.Mock())
        mock_controllers = mock.PropertyMock(
            side_effect=sushy.exceptions.MissingAttributeError)
        type(mock_storage).controllers = mock_controllers
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)
            mock_log.assert_called_once()

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_old_sushy(
            self, mock_get_system, mock_log):
        mock_storage = mock.Mock(spec=[])
        mock_storage.identity = "Storage 1"
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)
            mock_log.assert_called_once()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_old_sushy_oem(
            self, mock_get_system):
        mock_controller = mock.Mock()
        mock_controller.get_oem_extension.side_effect =\
            sushy.exceptions.ExtensionError
        mock_controllers_col = mock.Mock()
        mock_controllers_col.get_members.return_value = [mock_controller]
        mock_storage = mock.Mock(controllers=mock_controllers_col)
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(deploy_utils, 'get_async_step_return_state',
                       autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_delete_configuration_foreign_async(
            self, mock_get_system, mock_build_agent_options,
            mock_get_async_step_return_state, mock_convert):
        fake_oem_system = mock.Mock()
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system
        mock_get_system.return_value = fake_system
        task = mock.Mock(node=self.node, context=self.context)
        mock_return_state1 = mock.Mock()
        mock_return_state2 = mock.Mock()
        mock_get_async_step_return_state.return_value = mock_return_state2
        mock_oem_task1 = mock.Mock(
            job_type=sushy_oem_idrac.JOB_TYPE_RT_NO_REBOOT_CONF)
        mock_task1 = mock.Mock()
        mock_task1.get_oem_extension.return_value = mock_oem_task1
        mock_task_mon1 = mock.Mock(check_is_processing=True)
        mock_task_mon1.task_monitor_uri = '/TaskService/1'
        mock_task_mon1.get_task.return_value = mock_task1
        mock_oem_task2 = mock.Mock(job_type=sushy_oem_idrac.JOB_TYPE_RAID_CONF)
        mock_task2 = mock.Mock()
        mock_task2.get_oem_extension.return_value = mock_oem_task2
        mock_task_mon2 = mock.Mock(check_is_processing=False)
        mock_task_mon2.task_monitor_uri = '/TaskService/2'
        mock_task_mon2.get_task.return_value = mock_task2
        fake_oem_system.clear_foreign_config.return_value = [
            mock_task_mon1, mock_task_mon2]

        result = self.raid.post_delete_configuration(
            task, None, return_state=mock_return_state1)

        self.assertEqual(result, mock_return_state2)
        fake_oem_system.clear_foreign_config.assert_called_once()
        mock_build_agent_options.assert_called_once_with(task.node)
        mock_get_async_step_return_state.assert_called_once_with(task.node)
        mock_task_mon1.wait.assert_not_called()
        mock_task_mon2.wait.assert_not_called()
        mock_convert.assert_not_called()

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_delete_configuration_foreign_sync(
            self, mock_get_system, mock_convert):
        fake_oem_system = mock.Mock()
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system
        mock_get_system.return_value = fake_system
        task = mock.Mock(node=self.node, context=self.context)
        mock_return_state1 = mock.Mock()
        mock_oem_task1 = mock.Mock(
            job_type=sushy_oem_idrac.JOB_TYPE_RT_NO_REBOOT_CONF)
        mock_task1 = mock.Mock()
        mock_task1.get_oem_extension.return_value = mock_oem_task1
        mock_task_mon1 = mock.Mock(check_is_processing=True)
        mock_task_mon1.get_task.return_value = mock_task1
        mock_oem_task2 = mock.Mock(
            job_type=sushy_oem_idrac.JOB_TYPE_RT_NO_REBOOT_CONF)
        mock_task2 = mock.Mock()
        mock_task2.get_oem_extension.return_value = mock_oem_task2
        mock_task_mon2 = mock.Mock(check_is_processing=False)
        mock_task_mon2.get_task.return_value = mock_task2
        fake_oem_system.clear_foreign_config.return_value = [
            mock_task_mon1, mock_task_mon2]
        mock_convert_state = mock.Mock()
        mock_convert.return_value = mock_convert_state

        result = self.raid.post_delete_configuration(
            task, None, return_state=mock_return_state1)

        self.assertEqual(result, mock_convert_state)
        fake_oem_system.clear_foreign_config.assert_called_once()
        mock_task_mon1.wait.assert_called_once_with(CONF.drac.raid_job_timeout)
        mock_task_mon2.wait.assert_not_called()

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_clear_foreign_config', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_delete_configuration_no_subtasks(
            self, mock_get_system, mock_foreign, mock_convert):
        mock_foreign.return_value = False
        mock_convert.return_value = None
        task = mock.Mock(node=self.node, context=self.context)
        mock_return_state1 = mock.Mock()

        result = self.raid.post_delete_configuration(
            task, None, return_state=mock_return_state1)

        self.assertEqual(mock_return_state1, result)

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    def test__clear_foreign_config_attribute_error(self, mock_log):
        fake_oem_system = mock.Mock(spec=[])
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system

        result = drac_raid.DracRedfishRAID._clear_foreign_config(
            fake_system, mock.Mock())

        self.assertEqual(False, result)
        mock_log.assert_called_once()

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test__query_raid_tasks_status(self, mock_acquire):
        driver_internal_info = {'raid_task_monitor_uris': ['/TaskService/123']}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        mock_manager = mock.Mock()
        node_list = [(self.node.uuid, 'idrac', '', driver_internal_info)]
        mock_manager.iter_nodes.return_value = node_list
        task = mock.Mock(node=self.node,
                         driver=mock.Mock(raid=self.raid))
        mock_acquire.return_value = mock.MagicMock(
            __enter__=mock.MagicMock(return_value=task))
        self.raid._check_raid_tasks_status = mock.Mock()

        self.raid._query_raid_tasks_status(mock_manager, self.context)

        self.raid._check_raid_tasks_status.assert_called_once_with(
            task, ['/TaskService/123'])

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test__query_raid_tasks_status_no_task_monitor_url(self, mock_acquire):
        driver_internal_info = {'something': 'else'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        mock_manager = mock.Mock()
        node_list = [(self.node.uuid, 'idrac', '', driver_internal_info)]
        mock_manager.iter_nodes.return_value = node_list
        task = mock.Mock(node=self.node,
                         driver=mock.Mock(raid=self.raid))
        mock_acquire.return_value = mock.MagicMock(
            __enter__=mock.MagicMock(return_value=task))
        self.raid._check_raid_tasks_status = mock.Mock()

        self.raid._query_raid_tasks_status(mock_manager, self.context)

        self.raid._check_raid_tasks_status.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status(self, mock_get_task_monitor):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/123'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_message = mock.Mock()
        mock_message.message = 'Clear foreign config done'
        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = sushy.HEALTH_OK
        mock_config_task.messages = [mock_message]
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_get_task_monitor.return_value = mock_task_monitor

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/123'])

            self.raid._set_success.assert_called_once_with(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.raid._set_failed.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status_task_still_processing(
            self, mock_get_task_monitor):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/123'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_message = mock.Mock()
        mock_message.message = 'Clear foreign config done'
        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = sushy.HEALTH_OK
        mock_config_task.messages = [mock_message]
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_task_monitor2 = mock.Mock()
        mock_task_monitor2.is_processing = True
        mock_get_task_monitor.side_effect = [
            mock_task_monitor, mock_task_monitor2]

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()
        self.raid._substep_change_physical_disk_state_nonraid = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/123', '/TaskService/456'])

            (self.raid._substep_change_physical_disk_state_nonraid
                .assert_not_called())
            self.raid._set_success.assert_not_called()
            self.assertEqual(
                ['/TaskService/456'],
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.raid._set_failed.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status_task_failed(self, mock_get_task_monitor):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/123'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_message = mock.Mock()
        mock_message.message = 'Clear foreign config failed'
        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = 'Failed'
        mock_config_task.messages = [mock_message]
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_get_task_monitor.return_value = mock_task_monitor

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()
        self.raid._substep_change_physical_disk_state_nonraid = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/123'])

            (self.raid._substep_change_physical_disk_state_nonraid
                .assert_not_called())
            self.raid._set_success.assert_not_called()
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.raid._set_failed.assert_called_once()

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status_convert_controller(
            self, mock_get_task_monitor, mock_convert):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/1',
            'raid_config_substep': 'clear_foreign_config'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = sushy.HEALTH_OK
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_get_task_monitor.return_value = mock_task_monitor

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/1'])

            mock_convert.assert_called_once_with(task)
            self.raid._set_success.assert_not_called()
            self.raid._set_failed.assert_not_called()
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_config_substep'))

    @mock.patch.object(manager_utils, 'notify_conductor_resume_deploy',
                       autospec=True)
    @mock.patch.object(manager_utils, 'notify_conductor_resume_clean',
                       autospec=True)
    def test__set_success_clean(self, mock_notify_clean, mock_notify_deploy):
        self.node.clean_step = {'test': 'value'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_success(task)

            mock_notify_clean.assert_called_once_with(task)

    @mock.patch.object(manager_utils, 'notify_conductor_resume_deploy',
                       autospec=True)
    @mock.patch.object(manager_utils, 'notify_conductor_resume_clean',
                       autospec=True)
    def test__set_success_deploy(self, mock_notify_clean, mock_notify_deploy):
        self.node.clean_step = None
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_success(task)

            mock_notify_deploy.assert_called_once_with(task)

    @mock.patch.object(manager_utils, 'deploying_error_handler',
                       autospec=True)
    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    def test__set_failed_clean(self, mock_clean_handler, mock_deploy_handler):
        self.node.clean_step = {'test': 'value'}
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_failed(task, 'error', 'log message')

            mock_clean_handler.assert_called_once_with(
                task, 'error', 'log message')

    @mock.patch.object(manager_utils, 'deploying_error_handler',
                       autospec=True)
    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    def test__set_failed_deploy(self, mock_clean_handler, mock_deploy_handler):
        self.node.clean_step = None
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_failed(task, 'error', 'log message')

            mock_deploy_handler.assert_called_once_with(
                task, 'error', 'log message')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# TODO(stephenfin): Switch additionalProperties to False in a future version
index_request_query = {
    'type': 'object',
    'properties': {},
    'required': [],
    'additionalProperties': True,
}

index_response_body = {
    'type': 'object',
    'properties': {
        'shards': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'count': {'type': 'integer', 'minimum': 1},
                    'name': {'type': 'string'},
                },
                'required': ['count', 'name'],
                'additionalProperties': False,
            },
        },
    },
    'required': ['shards'],
    'additionalProperties': False,
}
